#include <htc.h>
#include "../hardware/buttons.h"
#include "../hardware/display.h"

// Stores the pattern to show on the display.
unsigned char display_data[4];

volatile unsigned char display_current_digit;
volatile unsigned char display_current_digit_mask;
volatile unsigned char display_current_segment_mask;

// Initialises the display pins and resets the segment scanning fields.
void display_init(void) {
	// Initialise the current digit/current segment counters.
	display_current_digit = 0;
	display_current_digit_mask = 0b00000001;
	display_current_segment_mask = 0b00000001;
}

// Switches the display off.
void display_off(void) {
	TMR0IE = 0;
	TMR0IF = 0;
	T0CS = 1;
	TMR0 = 0;
	TRISC &= 0xF0;
	PORTC &= 0xF0;
	TRISB = 0x00;
	PORTB = 0x00;
}

// Switches the display on.
void display_on(void) {
	PORTB = 0xFF;
	TRISB = 0x00;
	PORTC &= 0xF0;
	TRISC &= 0xF0;
	T0CS = 0;
	TMR0IE = 1;
	TMR0IF = 0;
}

// Switches the display off when it has finished scanning the entire display. This allows for flicker-free drawing when needed.
void display_off_synchronised(void) {
	while (display_current_digit != 3);
	while (display_current_segment_mask);
	display_off();
}

// Clear the display.
void display_clear(void) {
	display_data[0] = 0;
	display_data[1] = 0;
	display_data[2] = 0;
	display_data[3] = 0;
}

// Outputs a single segment. This function should be called in a loop.
void display_output_segment(void) {
	
	// Have we scanned all of the segments for the current digit?
	if (!display_current_segment_mask) {	
		PORTB = 0xFF;
		// Reset the segment mask and advance to the next digit.
		display_current_segment_mask = 0b00000001;
		++display_current_digit;
		// Is the digit still in range?
		if (display_current_digit < 4) {
			// Yes, so just shunt the mask along one.
			display_current_digit_mask <<= 1;
		} else {
			// No, so reset the digit index and mask.
			display_current_digit = 0;
			display_current_digit_mask = 0b00000001;
		}
		// Set the digit control lines.
		PORTC &= 0xF0;
		PORTC |= display_current_digit_mask;
	}

	// Output the segment.
	PORTB = (unsigned char)~(display_current_segment_mask & display_data[display_current_digit]);
	// Advance to the next segment.
	display_current_segment_mask <<= 1;
}

// Outputs a complete display frame.
void display_output_frame(void) {
	do {
		display_output_segment();
		__delay_us(100);
	} while (display_current_segment_mask && display_current_digit != 3);
}

// Converts a digit (in the range 0x0 to 0xF) to its segment pattern for a single space on the display.
unsigned char display_get_segments_for_digit(unsigned char digit) {
	const unsigned char digit_segments[] = {
		0b00111111,
		0b00000110,
		0b01011011,
		0b01001111,
		0b01100110,
		0b01101101,
		0b01111101,
		0b00000111,
		0b01111111,
		0b01101111,
		0b01110111,
		0b01111100,
		0b00111001,
		0b01011110,
		0b01111001,
		0b01110001,
	};
	return digit_segments[digit];
}

// Prints a 16-bit value as hex onto the display.
void display_print_hex_word(unsigned int value) {
	for (unsigned char d = 4; d > 0; --d) {
		display_data[d - 1] = display_get_segments_for_digit(value & 0xF);
		value >>= 4;
	}
}

// Prints a decimal (0000~9999) onto the display. If the MSB of the value is set, display padding zeroes.
void display_print_decimal(unsigned int value) {
	unsigned char hit_nonzero = (value >> 8) & 0x80;
	value &= 0x7FFF;
	if (value > 9999) {
		// Display "Err"
		display_set_display(0b00000000, 0b01111001, 0b01010000, 0b01010000);
	} else {
		// Split the number into is component digits.
		unsigned char digits[4];
		digits[0] = value / 1000;
		value -= digits[0] * 1000;
		digits[1] = value / 100;
		value -= digits[1] * 100;
		digits[2] = value / 10;
		value -= digits[2] * 10;
		digits[3] = value;
		// Copy the digits to the display.
		for (unsigned char digit = 0; digit < 4; ++digit) {
			if (digits[digit] || hit_nonzero || digit == 3) {
				display_data[digit] = display_get_segments_for_digit(digits[digit]);
				hit_nonzero = 1;
			} else {
				display_data[digit] = 0;
			}
		}
	}
}

// Prints a floating point value onto the display.
void display_print_float(float value) {
	
	unsigned char first_digit = 0;
	unsigned int factor = 1000;

	if (value < 0) {
		value = -value;
		++first_digit;
		factor = 100;
		display_data[0] = 0b01000000; // -
	}
	
	char si_prefix = 0;
	
	// Make the value < 1000, as the largest value we can show is 999 with an SI prefix.
	while (value >= factor) {
		value /= 1000.0f;
		++si_prefix;
	}
	// Make the value >= 1, to get as many digits after the decimal point on the display as possible.
	while (value > 0 && value < 1) {
		value *= 1000.0f;
		--si_prefix;
	}

	unsigned char decimal_place_position = 3;
	while (value < factor && decimal_place_position > first_digit) {
		value *= 10.0f;
		--decimal_place_position;
	}

	// Are we showing the prefix?
	unsigned char showing_prefix = (BUTTON_1 & BUTTON_2) && (si_prefix != 0);

	unsigned int i_value = (int)value;
	for (unsigned char digit = first_digit; digit < 4 - showing_prefix; ++digit) {
		unsigned char digit_value = i_value / factor;
		i_value -= digit_value * factor;
		factor /= 10;
		display_data[digit] = display_get_segments_for_digit(digit_value);
		if ((digit == decimal_place_position) && !(digit == 2 && showing_prefix)) display_data[digit] |= 0x80;
	}
	
	// Show the SI prefix if neither of the buttons are held down.
	if (showing_prefix) {
		const unsigned char prefix_chars[] = {
			0b01110011, // p
			0b01010100, // n
			0b00011100, // u
			0b01010101, // m
			0b00000000,
			0b01110110, // K
			0b00010101, // M
			0b00111101, // G
			0b01111000, // T
		};
		display_data[3] = prefix_chars[si_prefix + 4];
	}
}