#include <htc.h>

// Hardware support.
#include "../hardware/display.h"
#include "../hardware/buttons.h"
#include "../hardware/resistors.h"

// Settings.
#include "../main/settings.h"

// Probe modes.
#include "../modes/modes.h"
#include "../modes/probe.h"
#include "../modes/pulser.h"
#include "../modes/frequency.h"
#include "../modes/event_counter.h"
#include "../modes/voltmeter.h"
#include "../modes/capacitance.h"
#include "../modes/resistance.h"
#include "../modes/square_wave.h"
#include "../modes/temperature.h"
#include "../modes/one-wire_id.h"

// Utility.
#include "../utility/delay.h"

__CONFIG(HS & WDTDIS & LVPDIS & BORDIS);

void display_mode(void) {
	const unsigned char mode_strings[] = {
		0b01110011, 0b01010000, 0b01011100, 0b01111100, // Prob
		0b01110011, 0b00111110, 0b00111000, 0b01101101, // Puls
		0b01110001, 0b00110001, 0b01111011, 0b01100111, // Freq
		0b00000000, 0b00111001, 0b01010100, 0b01111000, // Cnt
		0b00111110, 0b01011100, 0b00111000, 0b01111000, // Volt
		0b01011110, 0b00000100, 0b01011100, 0b01011110, // Diod
		0b00000000, 0b00111001, 0b01011111, 0b01110011, // Cap
		0b00000000, 0b00110001, 0b01111011, 0b01101101, // Res
		0b00000000, 0b01101101, 0b01100111, 0b00110001, // Sqr
		0b00000110, 0b00101010, 0b00000100, 0b01011110, // 1Wid
		0b01111000, 0b01111011, 0b00010101, 0b01110011, // Temp
	};
	const unsigned char* src = mode_strings + (settings.mode * 4);
	for (unsigned char d = 0; d < 4; ++d) {
		display_data[d] = *(src++);
	}
}

int main(void) {

	display_init();
	display_on();
	ei();

	// Set all PORTA pins as digital inputs. 
	PCFG3 = 0;
	PCFG2 = 1;
	PCFG1 = 1;
	PCFG0 = 1;

	// Load the settings.
	settings_load();
	
	// Display the current mode.
	display_mode();

	// Delay.
	delay_ms(500);

	// Main program loop.
	for (;;) {
		// Is the "change mode" button held?
		while (!BUTTON_2) {
			// Do we need to advance mode?
			if (!BUTTON_1) {
				// Advance the mode.
				++settings.mode;
				if (settings.mode == PROBE_MODE_COUNT) settings.mode = 0;
				// Display the new mode.
				display_mode();
				// Delay a little for debouncing.
				delay_ms(10);
				button_1_wait_released();
			}
		}

		// We've just changed mode; save the settings.
		settings_save();

		// Run the selected mode.
		switch (settings.mode) {
			case PROBE_MODE_PROBE:
				probe();
				break;
			case PROBE_MODE_PULSER:
				pulser();
				break;
			case PROBE_MODE_FREQUENCY:
				frequency();
				break;
			case PROBE_MODE_EVENT_COUNTER:
				event_counter();
				break;
			case PROBE_MODE_VOLTMETER:
				voltmeter();
				break;
			case PROBE_MODE_DIODE_VOLTMETER:
				TRIS_R10K = 0;
				R10K = 1;
				voltmeter();
				break;
			case PROBE_MODE_CAPACITANCE:
				capacitance();
				break;
			case PROBE_MODE_RESISTANCE:
				resistance();
				break;
			case PROBE_MODE_SQUARE_WAVE:
				square_wave();
				break;
			case PROBE_MODE_1_WIRE_ID:
				one_wire_id();
				break;
			case PROBE_MODE_TEMPERATURE:
				temperature();
				break;
		}
		resistors_disconnect_all();
		
		// When we're here, it's because the exit key combination has been held.
		display_mode();
		// Save any settings that were saved when using the probe.
		settings_save();

		button_1_wait_released();
	}

}

