#include <htc.h>

#include "../main/isr.h"

#include "../hardware/display.h"
#include "../hardware/buttons.h"
#include "../hardware/resistors.h"

#include "../utility/delay.h"

// The time taken to take a reading of zero.
#define READING_ZERO (24.0f)
// The value of a known capacitance. I went through a bag of 10uF capacitors, measured each twice, then took the average.
#define READING_CALIBRATED (4588286.78f)
// The value of the capacitor used to calibrated the readings (in my case, 10uF).
#define CALIBRATION_CAPACITANCE (10e-6f)

void capacitance(void) {

	// Run the capacitance meter until both buttons are pressed.
	while (BUTTON_1 || BUTTON_2) {
	
		// Start by discharging the capacitor.
		R20 = 0;
		TRIS_R20 = 0;

		// Delay a while to discharge the capacitor.
		delay_ms(500);

		display_off();
	
		// Reset the timer.
		tmr1_overflow = 0;
		TMR1L = TMR1H = 0;
	
		// We'll want the timer to interrupts on overflow to maintain the MSB.
		TMR1IE = 1;
		TMR1IF = 0;
		PEIE = 1;
	
		// Charge up the capacitor through a resistor.
		R100K = 1;
		TRIS_R100K = 0;

		// Remove the discharge resistor to let the capacitor charge.
		TRIS_R20 = 1;
	
		// Start the timer.
		TMR1ON = 1;
		// Wait for the capacitor to charge.
		while (!R470);
		// Switch off the timer.
		TMR1ON = 0;
	
		// We don't want to generate any further interrupts.
		while (TMR1IF);
		TMR1IE = 0;
		PEIE = 0;
	
		// Disconnect the resistors.
		resistors_disconnect_all();
	
		// V=V0e^(-t/CR)
		// V/V0=e^(-t/CR)
		// ln(V/V0)=-t/CR
		// CRln(V/V0)=-t
		// CRln(V/V0)+t=0
		// Rln(V/V0)+t=-C
		// R and V/V0 are constants. C therefore has a linear relationship with t.
		
		// Get the time.
		float capacitance = (unsigned int)(TMR1L | (TMR1H << 8)) + ((float)tmr1_overflow * 65536.0f);

		// Convert it into an actual capacitance.
		capacitance = (capacitance - READING_ZERO) / (READING_CALIBRATED - READING_ZERO) * CALIBRATION_CAPACITANCE;	
		
		// Clamp to zero (negative capacitance is silly).
		if (capacitance < 0.0f) capacitance = 0.0f;

		// Switch the display back on.
		display_clear();
		display_on();

		// Uncomment out this block to display the raw reading on the display in hex for calibration.
		/*
		for (int d1 = 0; d1 < 100; ++d1) {
			display_print_hex_word(tmr1_overflow);
			delay_ms(10);
		}
		for (int d1 = 0; d1 < 100; ++d1) {
			display_print_hex_word(TMR1L | (TMR1H << 8));
			delay_ms(10);
		}
		*/

		// Display the result until button 1 is pressed again.
		while (BUTTON_1) {
			display_print_float(capacitance);
		}
		// If button 2 is pressed, bail out sharpish, else wait for button 1 to be released.
		if (BUTTON_2) {
			// Wait for button 1 to be released.
			delay_ms(10);
			while (!BUTTON_1);
			delay_ms(10);
		}
	}

}