#include <htc.h>

#include "../hardware/resistors.h"
#include "../hardware/buttons.h"

#define BUS_OUTPUT (TRIS_R20)
#define BUS_INPUT  (R20)

#define BUS_OUTPUT_LOW() BUS_OUTPUT=0
#define BUS_OUTPUT_HIGH() BUS_OUTPUT=1

#define BUS_OUTPUT_LOW_POWERED() BUS_INPUT=0
#define BUS_OUTPUT_HIGH_POWERED() BUS_INPUT=1

// Protocol delay times in microseconds.
// Names are taken from the Maxim documentation (AN126).
#define DELAY_A   (6)
#define DELAY_B  (64)
#define DELAY_C  (60)
#define DELAY_D  (10)
#define DELAY_E   (9)
#define DELAY_F  (55)
#define DELAY_G   (0)
#define DELAY_H (480)
#define DELAY_I  (70)
#define DELAY_J (410)

// Initialise the 1-Wire bus.
void one_wire_init(void) {
	// Enable a weak pull-up.
	TRIS_R10K = TRIS_R10KA = 0;
	R10K = R10KA = 1;
	// We'll drive the line low with a strong pull-down.
	BUS_INPUT = 0;
	BUS_OUTPUT = 1;
}

// Write a bit to the 1-Wire bus.
void one_wire_write_bit(unsigned char value) {
	if (value) {
		// Send a "1" bit.
		di();
		BUS_OUTPUT_LOW();
		__delay_us(DELAY_A);
		BUS_OUTPUT_HIGH();
		ei();
		__delay_us(DELAY_B);
	} else {
		// Send a "0" bit.
		di();
		BUS_OUTPUT_LOW();
		__delay_us(DELAY_C);
		BUS_OUTPUT_HIGH();
		ei();
		__delay_us(DELAY_D);
	}
}

// Write a byte to the 1-Wire bus.
void one_wire_write_byte(unsigned char value) {
	for (unsigned char bit_index = 0; bit_index < 8; ++bit_index) {
		if (value & 1) {
			// Send a "1" bit.
			di();
			BUS_OUTPUT_LOW();
			__delay_us(DELAY_A);
			BUS_OUTPUT_HIGH();
			ei();
			__delay_us(DELAY_B);
		} else {
			// Send a "0" bit.
			di();
			BUS_OUTPUT_LOW();
			__delay_us(DELAY_C);
			BUS_OUTPUT_HIGH();
			ei();
			__delay_us(DELAY_D);
		}
		value >>= 1;
	}
}

// Write a byte to the 1-Wire bus, and leave the bus in a powered state (strong pull-ups).
void one_wire_write_byte_powered(unsigned char value) {
	BUS_INPUT = 1;
	BUS_OUTPUT = 0;
	for (unsigned char bit_index = 0; bit_index < 8; ++bit_index) {
		if (value & 1) {
			// Send a "1" bit.
			di();
			BUS_OUTPUT_LOW_POWERED();
			__delay_us(DELAY_A);
			BUS_OUTPUT_HIGH_POWERED();
			ei();
			__delay_us(DELAY_B);
		} else {
			// Send a "0" bit.
			di();
			BUS_OUTPUT_LOW_POWERED();
			__delay_us(DELAY_C);
			BUS_OUTPUT_HIGH_POWERED();
			ei();
			__delay_us(DELAY_D);
		}
		value >>= 1;
	}
}

// Read a bit from the 1-Wire bus.
unsigned char one_wire_read_bit(void) {
	di();
	BUS_OUTPUT_LOW();
	__delay_us(DELAY_A);
	BUS_OUTPUT_HIGH();
	__delay_us(DELAY_E);
	unsigned char value = BUS_INPUT;
	ei();
	__delay_us(DELAY_F);
	return value;
}

// Read a byte from the 1-Wire bus.
unsigned char one_wire_read_byte(void) {
	unsigned char value = 0;
	for (unsigned char bit_index = 0; bit_index < 8; ++bit_index) {
		value >>= 1;
		di();
		BUS_OUTPUT_LOW();
		__delay_us(DELAY_A);
		BUS_OUTPUT_HIGH();
		__delay_us(DELAY_E);
		if (BUS_INPUT) value |= 0x80;
		ei();
		__delay_us(DELAY_F);
	}
	return value;
}

// Reset the 1-Wire bus slave devices. Returns 0 if a device asserts its presence, non-zero otherwise.
unsigned char one_wire_reset(void) {
	// Reset by pulling the line low for 820us.
	__delay_us(DELAY_G);
	BUS_OUTPUT_LOW();
	__delay_us(DELAY_H);
	BUS_OUTPUT_HIGH();
	__delay_us(DELAY_I);
	unsigned char presence = BUS_INPUT;	
	__delay_us(DELAY_J);
	return presence;
}

// Adapted from http://www.ccsinfo.com/forum/viewtopic.php?t=19520

unsigned char calculate_crc_byte(unsigned char crc, unsigned char value) { 
	unsigned char shift_reg, data_bit, sr_lsb, fb_bit, j; 
	
	shift_reg = crc; 
	for(unsigned char j = 0; j < 8; ++j) {
		data_bit = (value >> j) & 0x01;
		sr_lsb = shift_reg & 0x01;
		fb_bit = (data_bit ^ sr_lsb) & 0x01;
		shift_reg = shift_reg >> 1;
		if (fb_bit) shift_reg = shift_reg ^ 0x8c;
	} 
	return shift_reg; 
}

// Calculate the CRC for a block of data.
unsigned char one_wire_calculate_crc(unsigned char* data, unsigned char length) { 
	unsigned char crc = 0;
	while (length--) {
		crc = calculate_crc_byte(crc, *data);
		++data;
	}
	return crc;	
}